try:
    # noinspection PyPackageRequirements,PyUnresolvedReferences
    from twisted.internet import defer, reactor

    def async_sleep(seconds):
        """
        An asynchronous sleep function using twsited.

        Source: http://twistedmatrix.com/pipermail/twisted-python/2009-October/020788.html # noqa

        :type seconds: int
        """
        d = defer.Deferred()
        # noinspection PyUnresolvedReferences
        reactor.callLater(seconds, d.callback, seconds)
        return d

except ImportError:
    pass



from os import getenv


# noinspection SpellCheckingInspection
def listify(arg):
    """
    Simple utility method to ensure an argument provided is a list. If the
    provider argument is not an instance of `list`, then we return [arg], else
    arg is returned.

    :type arg: list
    :rtype: list
    """
    if isinstance(arg, (set, tuple)):
        # if it is a set or tuple make it a list
        return list(arg)
    if not isinstance(arg, list):
        return [arg]
    return arg


def resolve_setting(default, arg_value=None, env_var=None, config_value=None):
    """
    Resolves a setting for a configuration option. The winning value is chosen
    from multiple methods of configuration, in the following order of priority
    (top first):

    - Explicitly passed argument
    - Environment variable
    - Configuration file entry
    - Default

    :param arg_value: Explicitly passed value
    :param env_var: Environment variable name
    :type env_var: string or None
    :param config_value: Configuration entry
    :param default: Default value to if there are no overriding options
    :return: Configuration value
    """
    if arg_value is not None:
        return arg_value
    else:
        env_value = getenv(env_var)
        if env_value is not None:
            return env_value
        else:
            if config_value is not None:
                return config_value
            else:
                return default



import logging

TRACE = 5
logging.addLevelName(5, "TRACE")
logging.TRACE = 5


def trace(self, msg, *args, **kwargs):
    """
    Log 'msg % args' with severity 'TRACE'.

    To pass exception information, use the keyword argument exc_info with
    a true value, e.g.

    logger.trace("Houston, we have a %s", "thorny problem", exc_info=1)
    """
    if self.isEnabledFor(TRACE):
        self._log(TRACE, msg, args, **kwargs)


class TraceEnabledLogger(logging.Logger):
    trace = trace


logging.Logger.trace = trace

try:
    # noinspection PyProtectedMember
    LOGGING_LEVELS = logging._levelToName
except AttributeError:
    # noinspection PyProtectedMember
    LOGGING_LEVELS = logging._levelNames



from enum import Enum
from re import match

from cafeteria.compat import long

BYTES = 1
KB = 1024 * BYTES
MB = 1024 * KB
GB = 1024 * MB
TB = 1024 * GB


class MemoryUnit(Enum):
    BYTES = BYTES
    KB = KB
    MB = MB
    GB = GB
    TB = TB


class Memory(long):
    # noinspection PyInitNewSignature
    def __new__(cls, x, unit=None):
        if isinstance(x, str):
            units_regex = "|".join(MemoryUnit.__members__.keys())
            m = match(r"^(\d+) ?({})$".format(units_regex), x)
            if m is None:
                raise ValueError(
                    '{} requires am integer or a string in the format "<value>'
                    ' ({})"'.format(Memory.__class__.__name__, units_regex)
                )
            x = int(m.group(1)) * MemoryUnit.__members__.get(m.group(2)).value
        elif unit is None:
            raise ValueError("No unit provided.")
        else:
            x = x * unit.value
        # noinspection PyTypeChecker
        return super(Memory, cls).__new__(cls, x)



from logging import debug, exception, getLogger, root
from logging.config import dictConfig
from os import getenv
from os.path import isfile

from yaml import safe_load as load

from cafeteria.logging.trace import LOGGING_LEVELS
from cafeteria.patterns.mixins import ContextMixin


class LoggingManager(object):
    CONFIGFILE_ENV_KEY = "LOG_CFG"

    @classmethod
    def set_level(cls, level):
        """
        :raises: ValueError
        """
        level = (
            level
            if not isinstance(level, str)
            else int(LOGGING_LEVELS.get(level.upper(), level))
        )

        for handler in root.handlers:
            handler.setLevel(level)

        root.setLevel(level)

    @classmethod
    def load_config(cls, configfile="logging.yaml"):
        """
        :raises: ValueError
        """
        configfile = getenv(cls.CONFIGFILE_ENV_KEY, configfile)
        if isfile(configfile):
            with open(configfile, "r") as cf:
                # noinspection PyBroadException
                try:
                    dictConfig(load(cf))
                except ValueError:
                    debug("Learn to config foooo! Improper config at %s", configfile)
                except Exception:
                    exception("Something went wrong while reading %s.", configfile)
        else:
            raise ValueError("Invalid configfile specified: {}".format(configfile))


# noinspection PyPep8Naming
class LoggedObject(ContextMixin):
    def __new__(cls, *args, **kwargs):
        cls.logger = getLogger("{}.{}".format(cls.__module__, cls.__name__))
        """:type: cafeteria.logging.trace.TraceEnabledLogger"""
        cls.logger.trace("Instantiating %s.%s", cls.__module__, cls.__qualname__)
        return super(LoggedObject, cls).__new__(cls)

    def __enter__(self):
        self.logger.trace(
            "Entering context for %s.%s", self.__module__, self.__class__.__qualname__
        )
        return super(LoggedObject, self).__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.trace(
            "Exiting context for %s.%s", self.__module__, self.__class__.__qualname__
        )
        super(LoggedObject, self).__exit__(exc_type, exc_val, exc_tb)



from copy import deepcopy
from json import dumps, load, loads
from os.path import isfile

from cafeteria.patterns.borg import Borg


class AttributeDict(dict):
    """
    A dictionary implementation that allows for all keys to be used as an
    attribute. In this implementation we do proper get/setattr override here,
    no self.__dict__ mambo jumbo.
    """

    def __init__(self, *args, **kwargs):
        super(AttributeDict, self).__init__(*args, **kwargs)

    def __getattr__(self, item):
        if item in self:
            return self[item]
        raise AttributeError("Could not get attr: '{}' from '{}'".format(item, self))

    def __setattr__(self, key, value):
        self[key] = value


class DeepAttributeDict(AttributeDict):
    """
    A DeepAttributeDict is an AttributeDict of which dict objects at all depths
    are converted to DeepAttributeDict.
    """

    def __init__(self, *args, **kwargs):
        super(DeepAttributeDict, self).__init__(*args, **kwargs)
        self._deep_init()

    def _deep_init(self):
        for key, value in self.items():
            if isinstance(value, dict) and not isinstance(value, AttributeDict):
                self[key] = DeepAttributeDict(value)


class MergingDict(AttributeDict):
    """
    A MergingDict is an AttributeDict whose attribute/item values are always
    merged if the rvalue implements an update or append method. If the rvalue
    is not merge-able, it is simply replaced.
    """

    @property
    def disabled_types(self):
        return tuple()

    def replace(self, key, value):
        """
        Convenience method provided as a way to replace a value mapped by a
        key.This is required since a MergingDict always merges via assignment
        of item/attribute.

        :param key: Attribute name or item key to replace rvalue for.
        :type key: object
        :param value: The new value to assign.
        :type value: object
        :return:
        """
        super(MergingDict, self).__setitem__(key, value)

    def update(self, other=None, **kwargs):
        """
        A special update method to handle merging of dict objects. For all
        other iterable objects, we use the parent class update method. For
        other objects, we simply make use of the internal merging logic.

        :param other: An iterable object.
        :type other: dict or object
        :param kwargs: key/value pairs to update.
        :rtype: None
        """
        if other is not None:
            if isinstance(other, dict):
                for key in other:
                    self[key] = other[key]
            else:
                # noinspection PyTypeChecker
                super(MergingDict, self).update(other)

        for key in kwargs:
            self._merge(key, kwargs[key])

    def _merge_method(self, key):
        """
        Identify a merge compatible method available in self[key]. Currently we
        support 'update' and 'append'.

        :param key: Attribute name or item key
        :return: Method name usable to merge a value into the instance mapped
                by key
        :rtype: str
        """
        if key in self:
            obj = self[key]
            if isinstance(obj, self.disabled_types):
                return None
            for method in ["update", "append"]:
                if hasattr(obj, method):
                    return method
        return None

    def _merge(self, key, value):
        """
        Internal merge logic implementation to allow merging of values when
        setting attributes/items.

        :param key: Attribute name or item key
        :type key: str
        :param value: Value to set attribute/item as.
        :type value: object
        :rtype: None
        """
        method = self._merge_method(key)
        if method is not None and isinstance(self[key], type(value)):
            # strings are special, update methods like set.update looks for
            # iterables
            if method == "update" and isinstance(value, str):
                value = [value]
            if (
                method == "append"
                and isinstance(self[key], list)
                and isinstance(value, list)
            ):
                # if rvalue is a list and given object is a list, we expect all
                # values to be appended
                method = "extend"
            getattr(self[key], method)(value)
            return

        super(MergingDict, self).__setitem__(key, value)

    def __setitem__(self, key, value):
        self._merge(key, value)

    def __setattr__(self, key, value):
        self._merge(key, value)


class DeepMergingDict(MergingDict):
    """
    A DeepMergingDict is a MergingDict of which dict objects at all depths are
    converted to DeepMergingDicts.
    """

    def __init__(self, *args, **kwargs):
        super(DeepMergingDict, self).__init__(*args, **kwargs)
        self._deep_init()

    @staticmethod
    def _should_cast(value):
        return isinstance(value, dict) and not isinstance(value, MergingDict)

    def _deep_init(self):
        for key, value in self.items():
            if self._should_cast(value):
                self.replace(key, self.__class__(value))

    def replace(self, key, value):
        if self._should_cast(value):
            value = self.__class__(value)
        super(DeepMergingDict, self).replace(key, value)

    def update(self, other=None, **kwargs):
        if self._should_cast(other):
            other = self.__class__(other)
        super(DeepMergingDict, self).update(other, **kwargs)


class BorgDict(Borg, dict):
    """
    An dict implementing the Borg Pattern. This can be extended via
    inheritance. In this implementation the dict itself is not used. All
    actions are mapped to the Borg shared state.
    """

    def __init__(self, *args, **kwargs):
        super(BorgDict, self).__init__()
        self.update(*args, **kwargs)

    def update(self, *args, **kwargs):
        self.__dict__.update(*args, **kwargs)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def __delitem__(self, key):
        delattr(self, key)

    def __repr__(self):
        return self.__dict__.__repr__()

    def __str__(self):
        return self.__dict__.__str__()

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __contains__(self, k):
        return self.__dict__.__contains__(k)

    def keys(self):
        return self.__dict__.keys()

    def get(self, *args, **kwargs):
        return self.__dict__.get(*args, **kwargs)

    def pop(self, *args, **kwargs):
        return self.__dict__.pop(*args, **kwargs)


class JSONAttributeDict(AttributeDict):
    """
    :type source: str or dict or cafeteria.datastructs.dict.JSONAttributeDict
    """

    def __init__(self, source):
        super(JSONAttributeDict, self).__init__()

        try:
            self.update(loads(source) if isinstance(source, str) else deepcopy(source))
        except ValueError:
            if isfile(source):
                with open(source) as sf:
                    self.update(load(sf))
            else:
                raise ValueError(source)

    @property
    def pretty(self):
        return dumps(self, indent=2)

    def __str__(self):
        return self.pretty

    def __repr__(self):
        return self.pretty



class SessionManager(object):
    def __init__(self, factory, *args, **kwargs):
        self._kwargs = kwargs
        self._args = args
        self._factory = factory
        self.session = None

    def open(self):
        if self.session is None:
            self.session = self._factory(*self._args, **self._kwargs)

    def close(self):
        if self.session is not None:
            self.session.close()
        self.session = None

    def __enter__(self):
        self.open()
        return self.session

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



# noinspection PyUnresolvedReferences
from cafeteria.abc.compat import abstractclassmethod  # noqa F401


# noinspection PyPep8Naming,SpellCheckingInspection
class classproperty(property):
    """
    A decorator that behaves like @property except that operates
    on classes rather than instances.

    Original Implementation: sqlalchemy.util.langhelpers.classproperty
    """

    def __init__(self, fget, *arg, **kw):
        super(classproperty, self).__init__(fget, *arg, **kw)
        self.__doc__ = fget.__doc__

    # noinspection PyMethodOverriding,PyArgumentList,PyMethodParameters
    def __get__(desc, self, cls):
        return desc.fget(cls)



import logging
from typing import Any, Dict, Hashable, TypeVar, Union

logger = logging.getLogger(__name__)

D = TypeVar("D")


def get_by_path(d: Dict, *path: Hashable, default: D = None) -> Union[Any, D]:
    """
    Given a nested dict of dicts, traverse a given path and return the result or the default if it is not found.
    This is used as a replacement for the pattern
    >>> d.get("a", {}).get("b", {}).get("c", {}).get("d", 1)
    with
    >>> get_by_path(d, "a", "b", "c", "d", default=1)

    Does not traverse lists. Should be dict all the way down.
    """
    if len(path) == 0:
        raise ValueError("No path given")
    head, *tail = path
    logger.debug("head: %s, tail: %s", head, tail)
    if not tail:
        return d.get(head, default)
    try:
        return get_by_path(d[head], *tail, default=default)
    except KeyError:
        return default



from abc import ABCMeta


class AbstractClass(object):
    __metaclass__ = ABCMeta

    # noinspection PyUnresolvedReferences
    @classmethod
    def __subclasshook__(cls, other_class):
        return super(AbstractClass, cls).__subclasshook__(other_class) and all(
            any(x in B.__dict__ for B in other_class.__mro__)
            for x in cls.__abstractmethods__
        )



from re import match

try:
    long
except NameError:
    # noinspection PyShadowingBuiltins
    long = int


class BaseUnitClass(float):
    UNITS = {}

    # noinspection PyInitNewSignature
    def __new__(cls, x, unit=None):
        if isinstance(x, str):
            units_regex = "|".join(cls.UNITS.keys())
            m = match(r"^(\d+(.\d+)?) ?({})$".format(units_regex), x)
            if m is None:
                raise ValueError(
                    '{} requires number or a string in the format "<value> '
                    '({})"'.format(cls.__name__, units_regex)
                )
            x = float(m.group(1)) * cls.UNITS.get(m.group(3))
        elif unit is None:
            raise ValueError("No unit provided.")
        else:
            x = x * cls.UNITS[unit]
        return super(BaseUnitClass, cls).__new__(cls, x)

    def __getattr__(self, item):
        if item in self.UNITS:
            # if unit is known convert to unit
            result = self * 1.0 / self.UNITS[item]
            rounded = long(result)
            return result if result != rounded else rounded
        raise AttributeError("{} is not a valid conversion unit".format(item))



class BorgStateManager(object):
    """
    A special State Manager for Borg classes and child classes. This is what
    makes it possible for child classes to maintain their own state different
    to both parents, siblings and their own children.

    This itself implements the Borg pattern so that all its instances have a
    shared state.

    Each class state is mapped to the the hash of the class itself.
    """

    __shared_state = {}

    def __init__(self):
        self.__dict__ = self.__shared_state

    @classmethod
    def get_state(cls, clz):
        """
        Retrieve the state of a given Class.

        :param clz: types.ClassType
        :return: Class state.
        :rtype: dict
        """
        if clz not in cls.__shared_state:
            cls.__shared_state[clz] = (
                clz.init_state() if hasattr(clz, "init_state") else {}
            )
        return cls.__shared_state[clz]


class Borg(object):
    """
    A Borg pattern base class. Usable on its own or via inheritance. Uses
    `cafeteria.patterns.borg.BorgStateManager` internally to achieve state
    separation for children and grand children.

    See http://code.activestate.com/recipes/66531-singleton-we-dont-need-no-stinkin-singleton-the-bo/ for more # noqa
    information regarding the Borg Pattern.
    """

    def __init__(self):
        self.__dict__ = self._shared_state

    @classmethod
    def init_state(cls):
        return {}

    @property
    def _shared_state(self):
        return BorgStateManager.get_state(self.__class__)



from enum import Enum

from cafeteria.datastructs.units import BaseUnitClass


class DataMultiplier(Enum):
    k = 10**3
    M = 10**6
    G = 10**9
    T = 10**12
    P = 10**15
    E = 10**18
    Z = 10**21
    Y = 10**24
    Ki = 2**10
    Mi = 2**20
    Gi = 2**30
    Ti = 2**40
    Pi = 2**50
    Ei = 2**60


class DataBaseUnit(Enum):
    b = 1
    bit = 1
    B = 8
    byte = 8


class DataUnit(BaseUnitClass):
    """
    A data unit object internally stores the number of bits associated.
    Eg: DataUnit(1, 'byte') == 8
    """

    UNITS = {
        "{}{}".format(multiplier, base): DataMultiplier[multiplier].value
        * DataBaseUnit[base].value
        for multiplier in DataMultiplier.__members__
        for base in DataBaseUnit.__members__
    }
    UNITS.update({base: DataBaseUnit[base].value for base in DataBaseUnit.__members__})

    # noinspection PyInitNewSignature
    def __new__(cls, x, unit=None):
        if unit is None:
            # noinspection PyUnresolvedReferences
            unit = DataBaseUnit.bit.name
        return super(DataUnit, cls).__new__(cls, x, unit)


class DataRateUnit(DataUnit):
    """
    A data rate unit object internally stores the number bits per second.
    """

    UNITS = {
        "{}{}".format(unit, suffix): DataUnit.UNITS[unit]
        for unit in DataUnit.UNITS
        for suffix in ["/s", "ps"]
    }

    # noinspection PyInitNewSignature
    def __new__(cls, x, unit=None):
        if unit is None:
            # noinspection PyUnresolvedReferences
            unit = "{}ps".format(DataBaseUnit.bit.name)
        return super(DataUnit, cls).__new__(cls, x, unit)
